

function GetTextStream(input){
  var pos = 0;
  var line = 0;
  var col = 0;

  function peek(){
    return input.charAt(pos);
  }

  function next(){
    let c = input.charAt(pos);
    pos += 1;
    if (c === "\n"){
      line += 1;
      col = 0;
    } else {col += 1;}
    return c;
  }

  function eof(){
    return (input.charAt(pos) === "");
  }

  function getPos(){return pos;}
  function getLength(){return input.length;}

  function getLine(){return line;}
  function getCol(){return col;}

  function die(msg){
    throw new Error(msg + " Line: " + line + ", Col: " + col);
  }

  return {
    peek: peek,
    next: next,
    eof: eof,
    line: getLine,
    col: getCol,
    pos: getPos,
    length: getLength,
    die: die
  };
}


// ----------------------------------------------
// VALIDATORS
// ----------------------------------------------

function isWhiteSpace(c){
  return ("\t\n ".indexOf(c) >= 0);
}

function isStringStart(c){
  return (c === "\"" || c === "'");
}

function isNumType(c){
  return ("$%".indexOf(c) >= 0 || isDigit(c));
}

function isDigit(c){
  return /[0-9]/i.test(c);
}

function isHex(c){
  return /[0-9a-fA-F]/i.test(c);
}

function isBinary(c){
  return ("01".indexOf(c) >= 0);
}

function isLabelStart(c){
  return /[a-zA-Z_]/i.test(c);
}

function isLabel(c){
  return (isLabelStart(c) || isDigit(c));
}

function isPunctuation(c){
  return (",()".indexOf(c) >= 0);
}

function isOp(c){
  return ("=+-/*#<>!".indexOf(c) >= 0);
}


// -----------------------------------------------------------
// TOKENIZER CLASS
// -----------------------------------------------------------

class Tokenizer{

  constructor(op, input){
    if (!(op.hasOwnProperty("isCode") && op.hasOwnProperty("getCode")))
      throw new Error("OP Code definition object missing required method(s).");
    this.__stream = (input) ? GetTextStream(input) : GetTextStream("");
    this.__output = null;
    this.__op = op;
  }

  set input(i){
    if (typeof(i) === 'string'){
      this.__stream = GetTextStream(i);
      this.__output = null;
    }
  }
  get stream(){return this.__stream;}


  genTokenObject(type, val, line, col){
    return {
      type: type,
      val: val,
      line: (typeof(line) === 'number') ? line : this.__stream.line(),
      col: (typeof(col) === 'number') ? col : this.__stream.col()
    }
  }

// ----------------------------------------------
// Read Operations
// ----------------------------------------------

  skipComment(){
    this.readWhile((c)=>{return c != "\n";});
    this.__stream.next();
  }

  readHex(){
    this.__stream.next();
    var str = this.readWhile(isHex);
    return this.genTokenObject('number', parseInt(str, 16));
  }

  readBinary(){
    this.__stream.next();
    var str = this.readWhile(isBinary);
    return this.genTokenObject('number', parseInt(str, 2));
  }

  readNumber(){
    let c = this.__stream.peek();
    if (c === "$")
      return this.readHex();
    if (c === "%")
      return this.readBinary();

    var str = this.readWhile(isDigit);
    return this.genTokenObject('number', parseInt(str));
  }


  readString(end){
    var str = "";
    var escaped = false;
    var line = this.__stream.line();
    var col = this.__stream.col();
    this.__stream.next();
    while (!this.__stream.eof()){
      let c = this.__stream.next();
      if (escaped){
        str += c;
        escaped = false;
      } else if (c === "\\"){
        escaped = true;
      } else if (c === "\n"){
        this.__stream.die("String terminated early.");
      } else if (c === end){
        break;
      } else {
        str += c;
      }
    }
    return this.genTokenObject('string', str, line, col);
  }

  readDirective(){
    var str = this.__stream.next() + this.readWhile(isLabel);
    return this.genTokenObject('directive', str);
  }

  readLabel(){
    var str = this.readWhile(isLabel);
    if (this.__op.isCode(str.toUpperCase()))
      return this.genTokenObject('opcode', str.toUpperCase());
    return this.genTokenObject('label', str);
  }

  readWhile(validator){
    var str = "";
    while (!this.__stream.eof() && validator(this.__stream.peek()))
      str += this.__stream.next(); 
    return str;
  }

  nextToken(){
    this.readWhile(isWhiteSpace);
    if (this.__stream.eof()){return null;}
    let c = this.__stream.peek();
    if (c === ";"){
      this.skipComment();
      return this.nextToken();
    } else if (isStringStart(c)){
      return this.readString(c);
    } else if (isNumType(c)){
      return this.readNumber();
    } else if (isLabelStart(c)){
      return this.readLabel();
    } else if (c === "."){
      return this.readDirective();
    } else if (isPunctuation(c)){
      return this.genTokenObject('punc', this.__stream.next());
    } else if (isOp(c)){
      return this.genTokenObject('op', this.readWhile(isOp));
    }
    this.__stream.die("Unable to process character '" + c + "'.");
  }

  tokenize(input){
    if (input)
      this.input = input;

    if (this.__stream.length() > 0){
      if (this.__output === null && this.__stream.pos() === 0){
        this.__output = [];
        let t = this.nextToken();
        while (t !== null){
          this.__output.push(t);
          t = this.nextToken();
        }
      }
      return this.__output;
    }
    return [];
  }
}


module.exports = Tokenizer;




